@@ -1,17 +1,30 @@
 """KayPENTAX Disordered Voice Database Reader module"""
 
-__version__ = "0.4.0"
+__version__ = "0.6.0"
 
 import pandas as pd
 from os import path
 import numpy as np
-from glob import glob as _glob
+from glob import glob
 import re, operator
 import nspfile
+from typing import Literal, Tuple, List
+
+TaskType = Literal["AH", "RAINBOW"]
+
+DataField = Literal[
+    # fmt:off
+    "PAT_ID", "VISITDATE", "FILE AH", "NORM", "FILE RAINBOW", "AGE", "SEX",
+    "SMOKE", "NATLANG", "ORIGIN", "Fo", "To", "Fhi", "Flo", "STD", "PFR", "Fftr",
+    "Fatr", "Tsam", "Jita", "Jitt", "RAP", "PPQ", "sPPQ", "vFo", "ShdB", "Shim",
+    "APQ", "sAPQ", "vAm", "NHR", "VTI", "SPI", "FTRI", "ATRI", "DVB", "DSH",
+    "DUV", "NVB", "NSH", "NUV", "SEG", "PER",
+    # fmt:on
+]
 
 
 class KPDVDB:
-    def __init__(self, dbdir, remove_unknowns=False):
+    def __init__(self, dbdir: str, remove_unknowns: bool = False):
         """KPDVDB constructor
 
         :param dbdir: path to the cdrom drive or the directory hosting a copy of the database
@@ -36,29 +49,45 @@ class KPDVDB:
         * This function must be called at the beginning of each Python session
         * Database is loaded from the text file found at: <dbdir>/EXCEL50/TEXT/KAYCDALL.TXT
         * Only entries with NSP files are included
-        * PAT_ID of the entries without PAT_ID field uses the "FILE VOWEL 'AH'" field value
-        as the PAT_ID value
+        * Dataframe ID is set to the common prefix of AH and RAINBOW files
 
         """
 
         if self._dir == dbdir:
             return
 
-        ra_files = operator.iconcat(
-            *(
-                _glob(path.join(dbdir, vtype, "RAINBOW", "*.NSP"))
-                for vtype in ("NORM", "PATHOL")
-            )
+        # gather all files: name as key and NORM/PATHOL as value
+        file2df = lambda task: pd.DataFrame(
+            [
+                [s[-1][:5], s[-1], s[-3] == "NORM"]
+                for s in (
+                    p.split(path.sep)
+                    for p in operator.iconcat(
+                        *(
+                            glob(path.join(dbdir, vtype, task, "*.NSP"))
+                            for vtype in ("NORM", "PATHOL")
+                        )
+                    )
+                )
+            ],
+            columns=["ID", "FILE", "NORM"],
+        ).set_index("ID")
+
+        files = file2df("AH").join(
+            file2df("RAINBOW"), how="outer", lsuffix=" AH", rsuffix=" RAINBOW"
+        )
+        files["NORM"] = files["NORM AH"].where(
+            files["NORM AH"].notna(), files["NORM RAINBOW"]
         )
 
-        df = pd.read_table(
+        df_all = pd.read_table(
             path.join(dbdir, "EXCEL50", "TEXT", "KAYCDALL.TXT"),
             skipinitialspace=True,
             dtype={
                 # fmt: off
                 **{col: 'Int32' for col in ["AGE", "#", "NVB", "NSH", "NUV", "SEG", "PER"]},
-                **{col: "string" for col in ["PAT_ID", "FILE VOWEL 'AH'"]},
-                **{col: "category" for col in ["SEX", "DIAGNOSIS", "LOCATION", "NATLANG", "ORIGIN"]},
+                **{col: "string" for col in ["PAT_ID", "FILE VOWEL 'AH'", "DIAGNOSIS", "LOCATION"]},
+                **{col: "category" for col in ["SEX", "NATLANG", "ORIGIN"]},
                 **{col: float for col in ["Fo", "To", "Fhi", "Flo", "STD", "PFR", "Fftr", "Fatr", "Tsam", "Jita", 
                                         "Jitt", "RAP", "PQ", "sPPQ", "vFo", "ShdB", "Shim", "APQ", "sAPQ", 
                                         "vAm", "NHR", "VTI", "SPI", "FTRI", "ATRI", "DVB", "DSH", "DUV"]},
@@ -70,109 +99,88 @@ class KPDVDB:
             false_values=["N"],
         )
 
-        # add PAT_ID if missing (use unique ID based on the file name)
+        # drop the columns related to diagnosis
+        df = df_all.groupby("FILE VOWEL 'AH'")[
+            df_all.columns.drop(["#", "DIAGNOSIS", "LOCATION"])
+        ].first()
+
+        # reindex using the first 5 letters of the wav files
+        df.index = df.index.str[:5]
+        df.index.name = "ID"
+
+        # merge the rainbow files
+        df = df.merge(
+            files[["NORM", "FILE RAINBOW", "FILE AH"]],
+            "left" if remove_unknowns else "outer",
+            left_on="FILE VOWEL 'AH'",
+            right_on="FILE AH",
+        ).drop(columns=["FILE VOWEL 'AH'"])
+
         if remove_unknowns:
+            # drop all entries with unknown PAT_ID (incl. norm)
             df = df[df["PAT_ID"].notna()]
-        else:
-            tf = pd.isna(df["PAT_ID"])
-            for row in np.where(tf)[0]:
-                # id = df.at[row, "FILE VOWEL 'AH'"][:3]
-                # ids = df["PAT_ID"][df["PAT_ID"].str.startswith(id, na=False)].tolist()
-                # if len(ids):
-                #     for i in range(1000):
-                #         id_ = f"{id}{i:03}"
-                #         if id_ not in ids:
-                #             break
-                # else:
-                #     id_ = f"{id}000"
-                # df.at[row, "PAT_ID"] = id_
-                df.at[row, "PAT_ID"] = df.at[row, "FILE VOWEL 'AH'"]
+
+        # add PAT_ID if missing (use unique ID based on the file name)
+        df["ID"] = df["FILE AH"].str[:5]
+        tf = df["ID"].isna() & df["FILE RAINBOW"].notna()
+        df.loc[tf, "ID"] = df.loc[tf, "FILE RAINBOW"].str[:5]
+        df = df.set_index("ID").sort_index()
 
         # split dx
-        df_dx = df[["PAT_ID", "VISITDATE", "DIAGNOSIS", "LOCATION"]]
-        df.drop(columns=["#", "DIAGNOSIS", "LOCATION"], inplace=True)
-        df.drop_duplicates(subset=["PAT_ID", "VISITDATE"], inplace=True)
-        df.reset_index(inplace=True, drop=True)
+        df_dx = df_all.set_index("FILE VOWEL 'AH'")[["DIAGNOSIS", "LOCATION"]]
+        df_dx = df_dx[df_dx["DIAGNOSIS"].notna()]  # drop empty
 
-        # assign for rainbow
-        ra_names = set((path.basename(f) for f in ra_files))
-
-        # remove entries without PAT_ID or FILE VOWEL 'AH'
-        n = df.shape[0]
-        rnames = [None] * n
-        isnorms = np.empty(n, bool)
-        tf = np.zeros(n, bool)
-        for row in range(n):
-            aname = df.at[row, "FILE VOWEL 'AH'"]
-            if pd.isna(aname):
-                id = df.at[row, "PAT_ID"][:3]
-                rname = f"{id}1NRL.NSP"
-                isnorms[row] = isnorm = rname in ra_names
-                if not isnorm:
-                    day = df.at[row, "VISITDATE"].day
-                    rname = f"{id}{day:02}R.NSP"
-            else:
-                isnorms[row] = isnorm = aname.endswith("NAL.NSP")
-                rname = (
-                    re.sub(r"NAL\.NSP$", "NRL.NSP", aname)
-                    if isnorm
-                    else re.sub(r"AN\.NSP$", "R.NSP", aname)
-                )
-
-            if rname in ra_names:
-                rnames[row] = rname
-            else:
-                tf[row] = True
-
-        df.insert(3, "FILE RAINBOW", pd.Series(rnames, dtype="string"))
-        df.insert(3, "NORM", pd.Series(isnorms, dtype="boolean"))
-        df.drop(index=np.where(tf)[0], inplace=True)
-        df.reset_index(inplace=True, drop=True)
+        # update the df_dx to get the ID instead of the file name
+        df_dx = (
+            df_dx.merge(df["FILE AH"], "left", left_index=True, right_on="FILE AH")
+            .drop(columns="FILE AH")
+            .sort_index()
+        )
 
         self._dir = dbdir
         self._df = df
-        self._df_dx = df_dx[df_dx["DIAGNOSIS"].notna()]
+        self._df_dx = df_dx
         self._dx = None
 
-    def get_fields(self):
+    def get_fields(self) -> List[str]:
         """get list of all database fields
 
         :return: list of field names
-        :rtype: list(str)
+        :rtype: List[str]
         """
         return sorted([*self._df.columns.values, "DIAGNOSES"])
 
-    def get_sexes(self):
+    def get_sexes(self) -> List[str]:
         """get unique entries of SEX field
 
         :return: fixed list ["F", "M"]
-        :rtype: list(str)
+        :rtype: List[str]
         """
         return self._df["SEX"].cat.categories.tolist()
 
-    def get_natlangs(self):
+    def get_natlangs(self) -> List[str]:
         """get unique entries of NATLANGS field
 
         :return: list of subjects' native languages
-        :rtype: list(str)
+        :rtype: List[str]
         """
         return self._df["NATLANG"].cat.categories.tolist()
 
-    def get_origins(self):
+    def get_origins(self) -> List[str]:
         """get unique entries of ORIGIN field
 
         :return: list of subjects' races
-        :rtype: list(str)
+        :rtype: List[str]
         """
         return self._df["ORIGIN"].cat.categories.tolist()
 
-    def get_diagnoses(self, include_locations=False):
+    def get_diagnoses(self, include_locations=False) -> List[str]:
         """get unique entries of DIAGNOSIS field
 
         :param include_locations: True to also return diagnosis locations, defaults to False
         :type include_locations: bool, optional
         :return: list of diagnoses
-        :rtype: list(str)
+        :rtype: List[str]
         """
 
         if include_locations:
@@ -180,7 +188,6 @@ class KPDVDB:
         return self._df_dx["DIAGNOSIS"].cat.categories.tolist()
 
     def _get_dx_series(self):
-
         if self._dx is None:  # one-time operation
             self._dx = pd.Series(
                 [
@@ -197,16 +204,16 @@ class KPDVDB:
 
     def query(
         self,
-        subset=None,
-        include_diagnoses=False,
-        diagnoses_filter=None,
+        columns: List[DataField] = None,
+        include_diagnoses: bool = False,
+        diagnoses_filter: bool = None,
         **filters,
-    ):
+    ) -> pd.DataFrame:
         """query database
 
-        :param subset: database columns to return, defaults to None
-        :type subset: sequence of str, optional
-        :param include_diagnoses: True to include DIAGNOSES column. Ignored if subset or filters
+        :param columns: database columns to return, defaults to None
+        :type columns: List[DataField], optional
+        :param include_diagnoses: True to include DIAGNOSES column. Ignored if columns or filters
                                 specifies DIAGNOSES, defaults to False
         :type include_diagnoses: bool, optional
         :param diagnoses_filter: Function with the signature:
@@ -218,7 +225,7 @@ class KPDVDB:
         :return: query result
         :rtype: pandas.DataFrame
 
-        Valid values of `subset` argument (get_fields() + "MDVP")
+        Valid values of `columns` argument (get_fields() + "MDVP")
         ---------------------------------------------------------
 
         * All columns of the database specified in EXCEL50/TEXT/README.TXT Section 3.1
@@ -248,7 +255,7 @@ class KPDVDB:
         df = self._df.copy(deep=True)
         incl_dx = (
             include_diagnoses
-            or (subset is not None and "DIAGNOSES" in subset)
+            or (columns is not None and "DIAGNOSES" in columns)
             or "DIAGNOSES" in filters
         )
 
@@ -258,9 +265,13 @@ class KPDVDB:
         # apply the filters to reduce the rows
         for fcol, fcond in filters.items():
             try:
-                s = df[fcol]
+                if fcol == "ID":
+                    s = df.index
+                else:
+                    s = df[fcol]
             except:
                 raise ValueError(f"{fcol} is not a valid column label")
+
             if fcol == "DIAGNOSES":
                 fcond = set([fcond]) if isinstance(fcond, str) else set(fcond)
                 df = df[[len(fcond & set(v)) > 0 for v in s]]
@@ -284,34 +295,34 @@ class KPDVDB:
                 df.drop("DIAGNOSES", axis=1, inplace=True)
 
         # return only the selected columns
-        if subset is not None:
+        if columns is not None:
             try:
-                i = subset.index("MDVP")
+                i = columns.index("MDVP")
                 cols = df.columns
                 j = np.where(cols == "Fo")[0][0]
-                subset = [*subset[:i], *cols[j:].values, *subset[i + 1 :]]
+                columns = [*columns[:i], *cols[j:].values, *columns[i + 1 :]]
             except:
                 pass
             try:
-                df = df[subset]
+                df = df[columns]
             except:
                 ValueError(
-                    f'At least one label in the "subset" argument is invalid: {subset}'
+                    f'At least one label in the "columns" argument is invalid: {columns}'
                 )
 
         return df
 
     def get_files(
         self,
-        type,
-        auxdata_fields=None,
-        diagnoses_filter=None,
+        task: TaskType,
+        auxdata_fields: List[DataField] = None,
+        diagnoses_filter: List[str] = None,
         **filters,
-    ):
+    ) -> List[str] | Tuple[List[str], pd.DataFrame]:
         """get NSP filepaths
 
-        :param type: utterance type
-        :type type: "rainbow" or "ah"
+        :param task: utterance task
+        :type task: "rainbow" or "ah"
         :param auxdata_fields: names of auxiliary data fields to return, defaults to None
         :type auxdata_fields: sequence of str, optional
         :param diagnoses_filter: Function with the signature:
@@ -321,13 +332,14 @@ class KPDVDB:
         :param **filters: query conditions (values) for specific per-database columns (keys)
         :type **filters: dict
         :return: list of NSP files and optionally
-        :rtype: list(str) or tuple(list(str), pandas.DataFrame)
+        :rtype: List[str] or tuple(List[str], pandas.DataFrame)
 
         Valid values of `auxdata_fields` argument
         ---------------------------------
 
         * All columns of the database specified in EXCEL50/TEXT/README.TXT Section 3.1
         (except for "DIAGNOSIS" and "#")
+        * "ID" - recording ID returned by query()
         * "DIAGNOSES" - A list containing all the original "DIAGNOSIS" associated with the subject
         * "NORM" - True if normal data, False if pathological data
         * "MDVP" - Short-hand notation to include all the MDVP parameter measurements: from "Fo" to "PER"
@@ -351,17 +363,19 @@ class KPDVDB:
         try:
             col, subdir = {
                 "rainbow": ("FILE RAINBOW", "RAINBOW"),
-                "ah": ("FILE VOWEL 'AH'", "AH"),
-            }[type]
+                "ah": ("FILE AH", "AH"),
+            }[task]
         except:
-            raise ValueError(f'Unknown type: {type} (must be either "rainbow" or "ah")')
+            raise ValueError(
+                f'Unknown voice task: {task} (must be either "rainbow" or "ah")'
+            )
 
-        subset = [col, "NORM"]
+        columns = [col, "NORM"]
         if auxdata_fields is not None:
-            subset.extend(auxdata_fields)
+            columns.extend(auxdata_fields)
 
         df = self.query(
-            subset,
+            columns,
             diagnoses_filter=diagnoses_filter,
             **filters,
         )
@@ -380,17 +394,17 @@ class KPDVDB:
 
     def iter_data(
         self,
-        type,
-        channels=None,
-        auxdata_fields=None,
-        normalize=True,
-        diagnoses_filter=None,
+        task: TaskType,
+        channels: str | int | list = None,
+        auxdata_fields: List[DataField] = None,
+        normalize: bool = True,
+        diagnoses_filter: List[str] = None,
         **filters,
     ):
         """iterate over data samples
 
-        :param type: utterance type
-        :type type: "rainbow" or "ah"
+        :param task: utterance task
+        :type task: "rainbow" or "ah"
         :param channels: audio channels to read ('a', 'b', 0-1, or a sequence thereof),
                         defaults to None (all channels)
         :type channels: str, int, sequence, optional
@@ -429,6 +443,7 @@ class KPDVDB:
 
         * All columns of the database specified in EXCEL50/TEXT/README.TXT Section 3.1
         (except for "DIAGNOSIS" and "#")
+        * "ID" - recording ID returned by query()
         * "DIAGNOSES" - A list containing all the original "DIAGNOSIS" associated with the subject
         * "NORM" - True if normal data, False if pathological data
         * "MDVP" - Short-hand notation to include all the MDVP parameter measurements: from "Fo" to "PER"
@@ -450,7 +465,7 @@ class KPDVDB:
         """
 
         files = self.get_files(
-            type,
+            task,
             auxdata_fields,
             diagnoses_filter=diagnoses_filter,
             **filters,
@@ -461,7 +476,34 @@ class KPDVDB:
             files, auxdata = files
 
         for i, file in enumerate(files):
-            out = nspfile.read(file, channels)
-            if normalize:
-                out = (out[0], out[1] / 2.0 ** 15)
+            out = self._read_file(file, channels, normalize)
             yield (*out, auxdata.loc[i, :]) if hasaux else out
+
+    def read_data(
+        self,
+        id: int,
+        task: TaskType = None,
+        channels: str | int | list = None,
+        normalize: bool = True,
+    ) -> Tuple[int, np.array]:
+        """read audio data of one voice task of a recording session
+
+        :param id: recording id
+        :type id: int
+        :param task: voice task, defaults to None
+        :type task: TaskType, optional
+        :param channels: audio channel, defaults to None
+        :type channels: str | int | list, optional
+        :param normalize: True to normalize data between -1 and 1, defaults to True
+        :type normalize: bool, optional
+        :return: tuple of sampling rate in S/s and data samples numpy array
+        :rtype: tuple(int,np.array)
+        """
+        file = self.get_files(task, ID=id)[0]
+        return self._read_file(file, channels, normalize)
+
+    def _read_file(self, file, channels=None, normalize=True):
+        fs, x = nspfile.read(file, channels)
+        if normalize:
+            x = x / 2.0**15
+        return fs, x
